from magma.array import Array
from magma.circuit import Circuit
from magma.clock import AsyncReset, AsyncResetN, Clock
from magma.conversions import convertbit
from magma.interface import IO
from magma.passes import DefinitionPass, pass_lambda
from magma.t import In, Out, Direction
from magma.tuple import Tuple
from magma.wire import wire


class InsertWrapCasts(DefinitionPass):
    def define_wrap(self, wrap_type, in_type, out_type):
        class Wrap(Circuit):
            name = f"coreir_wrap{wrap_type}".replace("(", "").replace(")", "")
            io = IO(**{"in": In(in_type), "out": Out(out_type)})
            coreir_genargs = {"type": wrap_type}
            coreir_name = "wrap"
            coreir_lib = "coreir"

            def simulate(self, value_store, state_store):
                input_val = value_store.get_value(getattr(self, "in"))
                value_store.set_value(self.out, input_val)
        return Wrap

    def wrap_if_named_type(self, port, definition):
        if isinstance(port, Tuple):
            wrapped = False
            for t in port:
                wrapped |= self.wrap_if_named_type(t, definition)
            return wrapped
        if isinstance(port, Array):
            wrapped = self.wrap_if_named_type(port[0], definition)
            if not wrapped:
                return False
            # TODO: Magma doesn't support length zero array, so slicing a
            # length 1 array off the end doesn't work as expected in normal
            # Python, so we explicilty slice port.ts
            for t in port.ts[1:]:
                self.wrap_if_named_type(t, definition)
            return True
        if not port.driven():
            return False
        value = port.value()
        if not (isinstance(port, (AsyncReset, AsyncResetN, Clock)) or
                isinstance(value, (AsyncReset, AsyncResetN, Clock))):
            return self.wrap_if_named_type(value, definition)
        undirected_t = type(port).qualify(Direction.Undirected)
        if issubclass(type(value), undirected_t):
            return self.wrap_if_named_type(value, definition)
        port_is_clock_type = isinstance(port, (AsyncReset, AsyncResetN, Clock))
        if port_is_clock_type:
            T = Out(type(port))
        else:
            T = In(type(value))

        with definition.open():
            port.unwire(value)
            in_type, out_type = type(port), type(value)
            if not port_is_clock_type:
                # value is the clock type, flip wrap types
                out_type, in_type = in_type, out_type
            inst = self.define_wrap(T, in_type, out_type)()
            wire(convertbit(value, in_type), getattr(inst, "in"))
            wire(inst.out, convertbit(port, out_type))
        return True

    def __call__(self, definition):
        # copy, because wrapping might add instances
        instances = definition.instances[:]
        for instance in instances:
            if type(instance).coreir_name == "wrap" or \
                    type(instance).coreir_name == "unwrap":
                continue
            for port in instance.interface.ports.values():
                self.wrap_if_named_type(port, definition)
        for port in definition.interface.ports.values():
            self.wrap_if_named_type(port, definition)


insert_wrap_casts = pass_lambda(InsertWrapCasts)



from .clock import (AbstractReset, AsyncReset, AsyncResetN, Clock, Enable,
                    Reset, ResetN, get_reset_args)
from .interface import IO
from .t import In


def ClockIO(has_enable=False, has_reset=False, has_resetn=False, has_ce=False,
            has_async_reset=False, has_async_resetn=False):
    args = {'CLK': In(Clock)}
    has_enable |= has_ce
    if has_enable:
        args['CE'] = In(Enable)
    if has_reset:
        args['RESET'] = In(Reset)
    if has_resetn:
        args['RESETN'] = In(ResetN)
    if has_async_reset:
        args['ASYNCRESET'] = In(AsyncReset)
    if has_async_resetn:
        args['ASYNCRESETN'] = In(AsyncResetN)
    return IO(**args)


def gen_clock_io(reset_type: AbstractReset = None, has_enable: bool = False):
    (
        has_async_reset, has_async_resetn, has_reset, has_resetn
    ) = get_reset_args(reset_type)

    return ClockIO(has_enable=has_enable,
                   has_async_reset=has_async_reset,
                   has_async_resetn=has_async_resetn,
                   has_reset=has_reset,
                   has_resetn=has_resetn)



from collections import namedtuple

class Scope:
    def __init__(self, **kwargs):
        self.inst = None
        self.parent = None
        if 'parent' in kwargs:
            self.parent = kwargs['parent']
            self.val = self.parent.val
            if self.val != "/":
                self.val += '/'
        else:
            self.val = '/'

        if 'instance' in kwargs:
            self.inst = kwargs['instance']
            self.val += type(self.inst).__name__ + '.' + self.inst.name

    def inst_scope():
        return self.inst is not None

    def value(self):
        return self.val

    def __eq__(self, rhs):
        return self.val == rhs.val

    def __hash__(self):
        return hash(self.val)

QualifiedBit = namedtuple('QualifiedBit', ['bit', 'scope'])
QualifiedInstance = namedtuple('QualifiedInstance', ['instance', 'scope'])



import collections
import collections.abc
from functools import wraps, partial
import warnings


class Stack:
    """Lightweight wrapper over builtin lists to provide a stack interface"""
    def __init__(self):
        self._stack = []

    def push(self, value):
        self._stack.append(value)

    def pop(self):
        return self._stack.pop()

    def peek(self):
        return self._stack[-1]


class _Ref(object):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value is other.value

    def __hash__(self):
        return id(self.value)


class _IdentitySetBase(collections.abc.MutableSet):
    def __init__(self, refs):
        self.refs = refs

    def __contains__(self, elem):
        return _Ref(elem) in self.refs

    def __iter__(self):
        return (ref.value for ref in self.refs)

    def __len__(self):
        return len(self.refs)

    def add(self, elem):
        self.refs.add(_Ref(elem))

    def discard(self, elem):
        raise NotImplementedError()

    def remove(self, elem):
        raise NotImplementedError()

    def pop(self):
        raise NotImplementedError()

    def clear(self):
        self.refs.clear()

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, list(self))


class IdentitySet(_IdentitySetBase):
    def __init__(self, items=()):
        refs = set(map(_Ref, items))
        super().__init__(refs)


class OrderedIdentitySet(_IdentitySetBase):
    def __init__(self, items=()):
        # NOTE(rsetaluri): We use collections.OrderedDict to mimic an ordered
        # set, to avoid implementing a custom ordered set or import one, since
        # it is not natively supported.
        refs = map(lambda x: (x, None), map(_Ref, items))
        refs = collections.OrderedDict(refs)
        super().__init__(refs)

    def add(self, elem):
        self.refs[_Ref(elem)] = None


def deprecated(func=None, *, msg=None):
    """
    Inspired by https://pybit.es/decorator-optional-argument.html.

    Marks a function as deprecated. @msg is an optional parameter to override
    the default warning.

    Examples:

        @deprecated
        def foo(...): ...

        @deprecated('Don't use bar!')
        def bar(...): ...
    """
    if func is None:
        return partial(deprecated, msg=msg)
    if msg is None:
        msg = f"{func.__name__} is deprecated"

    @wraps(func)
    def _wrapper(*args, **kwargs):
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)

    return _wrapper


def setattrs(obj, dct, pred=None):
    for k, v in dct.items():
        if pred is None or pred(k, v):
            setattr(obj, k, v)


class ParamDict(dict):
    """
    Hashable dictionary for simple key: value parameters
    """
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


def is_int(value):
    try:
        int(value)
    except (TypeError, ValueError):
        return False
    return True



from collections import OrderedDict
from ..bit import Bit, Digital
from ..array import Array
from ..clock import wiredefaultclock
from ..compiler import make_compiler
from .verilog import find

def get_type(port):
    if isinstance(port, Array):
        width = port.N
    else:
        assert isinstance(port, Bit)
        width = 1
    return "UInt<{}>".format(width)

def get_name(port):
    if isinstance(port, Digital) and port.const():
        return ("UInt<1>(\"h1\")"
                if port is type(port).VCC
                else "UInt<1>(\"h0\")")

    if isinstance(port, Array):
        if not port.iswhole():
            # the sequence of values is concantenated
            port = [get_name(i) for i in port.ts]
            port.reverse()
            if len(port) == 1:  # FIXME: Hack to make single length bit arrays work
                return port[0]
            return '{' + ','.join(port) + '}'
    assert not port.anon()
    return port.name.qualifiedname(sep="_")

def compileinstance(instance):
    instance_name = str(instance.name)
    s = "inst {} of {}\n".format(instance_name, str(instance.__class__.__name__))
    for name, port in instance.interface.ports.items():
        if port.is_input():
            value = port.value()
            if not value:
                print('Warning (firrtl): input', str(port), 'not connected to an output')
                value = port
        else:
            value = port
        # if isinstance(value, Array):
        #     for index, subport in enumerate(value.ts):
        #         s += "{}.{}[{}] <= {}\n".format(instance_name, name, index, get_name(subport))
        # else:
        value_name = get_name(value)
        if port.is_input():
            s += "{}.{} <= {}\n".format(instance_name, name, value_name)
        else:
            s += "{} <= {}.{}\n".format(value_name, instance_name, name)
    return s


def compile_primitive(instance):
    op = instance.firrtl_op
    instance_name = str(instance.name)
    # s = "inst {} of {}\n".format(instance_name, str(instance.__class__.__name__))
    outputs = instance.interface.outputs()
    assert len(outputs) == 1, "FIRRTL primitive should only have one output"
    port = outputs[0]
    args = []
    for name, port in instance.interface.ports.items():
        if port.is_input():
            value = port.value()
            if not value:
                print('Warning (firrtl): input', str(port), 'not connected to an output')
                value = port
            args.append(get_name(value))
    return "{} <= {}({})\n".format(get_name(port), instance.firrtl_op, ", ".join(args))

def compiledefinition(cls):

    # for now only allow Bit or Array(n, Bit)
    for name, port in cls.interface.ports.items():
        if isinstance(port, ArrayKind):
            if not isinstance(port.T, BitKind):
                print('Error: Argument', port, 'must be a an Array(n,Bit)')

    s = 'module {} :\n'.format(cls.__name__)
    for name, port in cls.interface.ports.items():
        if port.is_input():
            direction = "output"
        elif port.is_output():
            direction = "input"
        else:
            raise NotImplementedError()  # Does FIRRTL have an inout?
        s += '{} {} : {}\n'.format(direction, name, get_type(port))
    s += "\n"  # Newline after port declaration for readability

    import re
    if cls.firrtl:
        s += cls.firrtl + '\n'
    else:
        # declare a wire for each instance output
        for instance in cls.instances:
            for port in instance.interface.ports.values():
                if port.is_output():
                    s += 'wire {} : {}\n'.format(get_name(port), get_type(port))

        #print('compile instances')
        # emit the structured verilog for each instance
        for instance in cls.instances:
            if getattr(instance, "firrtl_op", False):
                s += compile_primitive(instance)
            else:
                wiredefaultclock(cls, instance)
                s += compileinstance(instance)

        # assign to module output arguments
        for input in cls.interface.inputs():
            output = input.value()
            if output:
                iname = get_name(input)
                oname = get_name(output)
                s += '%s <= %s\n' % (iname, oname)

    return "\n  ".join(s.splitlines())


def compile(main):
    defn = find(main,OrderedDict())

    code = 'circuit main :\n'
    for k, v in defn.items():
         print('compiling', k)
         code += "  " + "\n  ".join(compiledefinition(v).splitlines()) + '\n'
    return code


FirrtlCompiler = make_compiler("fir", compile)



from inspect import getouterframes, currentframe
from pathlib import PurePath
from .backend import verilog, blif, firrtl, dot
from .compiler import Compiler
from .config import get_compile_dir
from .uniquification import uniquification_pass, UniquificationMode
from .passes.clock import WireClockPass
from .passes.drive_undriven import DriveUndrivenPass
from .passes.terminate_unused import TerminateUnusedPass
from magma.backend.coreir.coreir_compiler import CoreIRCompiler
from magma.bind import BindPass
from magma.inline_verilog import ProcessInlineVerilogPass

__all__ = ["compile"]


def _make_compiler(output, main, basename, opts):
    if output == "verilog":
        return verilog.VerilogCompiler(main, basename)
    if output == "blif":
        return blif.BlifCompiler(main, basename)
    if output == "firrtl":
        return firrtl.FirrtlCompiler(main, basename)
    if output == "dot":
        return dot.DotCompiler(main, basename)
    if output == "coreir":
        return CoreIRCompiler(main, basename, opts)
    if output == "coreir-verilog":
        opts["output_verilog"] = True
        return CoreIRCompiler(main, basename, opts)
    raise NotImplementedError(f"Backend '{output}' not supported")


def _get_basename(basename):
    if get_compile_dir() == "callee_file_dir":
        callee_filename = getouterframes(currentframe())[2][1]
        callee_dir = PurePath(callee_filename).parent
        return str(callee_dir.joinpath(basename))
    return basename


def compile(basename, main, output="coreir-verilog", **kwargs):
    if hasattr(main, "circuit_definition"):
        main = main.circuit_definition
    basename = _get_basename(basename)
    opts = kwargs.copy()
    compiler = _make_compiler(output, main, basename, opts)

    # Default behavior is to perform uniquification, but can be overriden.
    uniquification_pass(main, opts.get("uniquify", "UNIQUIFY"))

    # Steps to process inline verilog generation. Required to be run after
    # uniquification.
    ProcessInlineVerilogPass(main).run()

    # Bind after uniquification so the bind logic works on unique modules.
    BindPass(main, compile, opts.get("user_namespace")).run()

    if opts.get("drive_undriven", False):
        DriveUndrivenPass(main).run()
    if opts.get("terminate_unused", False):
        TerminateUnusedPass(main).run()

    result = compiler.compile()
    result = {} if result is None else result

    if hasattr(main, "fpga"):
        main.fpga.constraints(basename)

    return result



import ast
from copy import deepcopy


class SymbolReplacer(ast.NodeTransformer):
    def __init__(self, symbol_table, ctx):
        self.symbol_table = symbol_table
        self.ctx = ctx

    def visit_Name(self, node):
        if node.id in self.symbol_table:
            if self.ctx is None:
                new_node = deepcopy(self.symbol_table[node.id])
                new_node.ctx = node.ctx
                return new_node
            elif isinstance(node.ctx, self.ctx):
                return deepcopy(self.symbol_table[node.id])
        return node


def replace_symbols(tree, symbol_table, ctx=None):
    return SymbolReplacer(symbol_table, ctx).visit(tree)



import dataclasses
from typing import List, Optional

from magma.array import Array
from magma.bits import Bits
from magma.clock import Clock
from magma.circuit import Circuit, CircuitBuilder, _DefinitionContextManager
from magma.digital import Digital
from magma.ref import AnonRef, ArrayRef, DefnRef, InstRef, TupleRef
from magma.t import In
from magma.tuple import Tuple


@dataclasses.dataclass(frozen=True)
class _CompileGuardState:
    ckt: CircuitBuilder
    ctx_mgr: _DefinitionContextManager


class _CompileGuardBuilder(CircuitBuilder):
    def __init__(self, name, cond):
        super().__init__(name)
        self._cond = cond
        self._system_types_added = set()
        self._set_inst_attr("coreir_metadata", {"compile_guard": cond})
        self._port_index = 0
        self._pre_finalized = False

    def _new_port_name(self):
        name = f"port_{self._port_index}"
        self._port_index += 1
        return name

    def _rewire(self, port, value):
        new_port_name = self._new_port_name()
        T = type(value).undirected_t
        self._add_port(new_port_name, In(T))
        new_port = self._port(new_port_name)
        port.unwire(value)
        port @= new_port
        external = getattr(self, new_port_name)
        external @= value

    def pre_finalize(self):
        if self._pre_finalized:
            raise Exception("Can not call pre_finalize multiple times")
        for inst in self._instances:
            self._process_instance(inst)
        self._pre_finalized = True

    def _process_instance(self, inst):
        for port in inst.interface.inputs(include_clocks=True):
            self._process_port(port)

    def _process_port(self, port, ref=None, kwargs=None):
        value = port.value()
        if ref is None:
            if value is None:
                self._process_undriven(port)
                return
            ref = value.name
        if isinstance(ref, AnonRef):
            if value.const():
                return
            # TODO(rsetaluri): Implement valid anon. values.
            raise NotImplementedError()
        if isinstance(ref, DefnRef):
            self._rewire(port, value)
            return
        if isinstance(ref, InstRef):
            internal = any(ref.inst is inst for inst in self._instances)
            if internal:
                return
            self._rewire(port, value)
            return
        if isinstance(ref, TupleRef):
            # TODO(rsetaluri): Fix port name collision.
            self._process_port(port, ref=ref.tuple.name)
            return
        if isinstance(ref, ArrayRef):
            # TODO(rsetaluri): Fix port name collision.
            self._process_port(port, ref=ref.array.name)
            return
        # TODO(rsetaluri): Do the rest of these.
        raise NotImplementedError(ref, type(ref))

    def _process_undriven(self, port):
        # TODO(rsetaluri): Add other system types
        if not isinstance(port, Clock):
            return
        T = type(port).undirected_t
        if T in self._system_types_added:
            return
        name = str(port.name)
        self._add_port(name, In(T))
        self._system_types_added.add(T)


class _CompileGuard:
    __index = 0

    def __init__(self, cond: str,
                 defn_name: Optional[str],
                 inst_name: Optional[str]):
        self._cond = cond
        self._defn_name = defn_name
        self._inst_name = inst_name
        self._state = None

    @staticmethod
    def _new_name():
        index = _CompileGuard.__index
        _CompileGuard.__index += 1
        return f"CompileGuardCircuit_{index}"

    def __enter__(self):
        if self._state is not None:
            raise Exception("Can not enter compile guard multiple times")
        assert self._state is None
        if self._defn_name is None:
            self._defn_name = _CompileGuard._new_name()
        if self._state is None:
            ckt = _CompileGuardBuilder(self._defn_name, self._cond)
            if self._inst_name is None:
                ckt.set_instance_name(self._inst_name)
            ctx_mgr = _DefinitionContextManager(ckt._context)
            self._state = _CompileGuardState(ckt, ctx_mgr)
        self._state.ctx_mgr.__enter__()

    def __exit__(self, typ, value, traceback):
        self._state.ctx_mgr.__exit__(typ, value, traceback)
        self._state.ckt.pre_finalize()


def compile_guard(cond: str,
                  defn_name: Optional[str] = None,
                  inst_name: Optional[str] = None):
    return _CompileGuard(cond, defn_name, inst_name)



from abc import ABC, abstractmethod
from copy import copy
import json
import logging
import os
from magma.digital import Digital
from magma.array import Array
from magma.bits import Bits
from coreir import Wireable
from magma.backend.coreir.coreir_utils import (
    attach_debug_info, check_magma_interface, constant_to_value, get_inst_args,
    get_module_of_inst, magma_interface_to_coreir_module_type,
    magma_port_to_coreir_port, make_cparams, map_genarg,
    magma_name_to_coreir_select, Slice)
from magma.interface import InterfaceKind
from magma.is_definition import isdefinition
from magma.logging import root_logger
from magma.passes import InstanceGraphPass
from magma.tuple import Tuple
from magma.backend.util import get_codegen_debug_info
from magma.clock import (wire_default_clock, is_clock_or_nested_clock,
                         get_default_clocks)
from magma.config import get_debug_mode
from magma.protocol_type import MagmaProtocol, MagmaProtocolMeta
from magma.ref import PortViewRef, ArrayRef


# NOTE(rsetaluri): We do not need to set the level of this logger since it has
# already been done in backend/coreir/coreir_backend.py.
_logger = root_logger().getChild("coreir_backend")


def _make_unconnected_error_str(port):
    error_str = port.debug_name
    if port.trace() is not None:
        error_str += ": Connected"
    elif isinstance(port, (Tuple, Array)):
        child_str = ""
        for child in port:
            child = _make_unconnected_error_str(child)
            child = "\n    ".join(child.splitlines())
            child_str += f"\n    {child}"
        if "Connected" not in child_str:
            # Handle case when no children are connected (simplify)
            error_str += ": Unconnected"
        else:
            error_str += child_str
    elif port.trace() is None:
        error_str += ": Unconnected"
    return error_str


def _collect_drivers(value):
    """
    Iterate over value to collect the child drivers, packing slices together
    """
    drivers = []
    start_idx = 0
    for i in range(1, len(value)):
        # If the next value item is not a reference to an array of bits where
        # the array matches the previous item and the index is incremented by
        # one, append the current slice to drivers (may introduce slices of
        # length 1)
        if not (
            isinstance(value[i].name, ArrayRef) and
            issubclass(value[i].name.array.T, Digital) and
            isinstance(value[i - 1].name, ArrayRef) and
            value[i].name.array is value[i - 1].name.array and
            value[i].name.index == value[i - 1].name.index + 1
        ):
            drivers.append(value[start_idx:i])
            start_idx = i
    drivers.append(value[start_idx:])
    return drivers


def _unwrap(x):
    if isinstance(x, MagmaProtocol):
        return x._get_magma_value_()
    if isinstance(x, MagmaProtocolMeta):
        return x._to_magma_()
    return x


class TransformerBase(ABC):
    def __init__(self, backend, opts):
        self.backend = backend
        self.opts = opts
        self.ran = False
        self._children = None

    def run(self):
        if self.ran:
            raise RuntimeError("Can only run transformer once")
        self._children = self.children()
        for child in self._children:
            child.run()
        self.run_self()
        self.ran = True

    @abstractmethod
    def children(self):
        raise NotImplementedError()

    def run_self(self):
        pass


class LeafTransformer(TransformerBase):
    def children(self):
        return []


class DefnOrDeclTransformer(TransformerBase):
    def __init__(self, backend, opts, defn_or_decl):
        super().__init__(backend, opts)
        self.defn_or_decl = defn_or_decl
        self.coreir_module = None

    def children(self):
        try:
            coreir_module = self.backend.get_module(self.defn_or_decl)
            _logger.debug(f"{self.defn_or_decl} already compiled, skipping")
            self.coreir_module = coreir_module
            return []
        except KeyError:
            pass
        if not isdefinition(self.defn_or_decl):
            return [DeclarationTransformer(self.backend,
                                           self.opts,
                                           self.defn_or_decl)]
        wrapped = getattr(self.defn_or_decl, "wrappedModule", None)
        if wrapped and wrapped.context is self.backend.context:
            return [WrappedTransformer(self.backend,
                                       self.opts,
                                       self.defn_or_decl)]
        return [DefinitionTransformer(self.backend,
                                      self.opts,
                                      self.defn_or_decl)]

    def run_self(self):
        if self.coreir_module:
            return
        self.coreir_module = self._children[0].coreir_module
        self.backend.add_module(self.defn_or_decl, self.coreir_module)
        if isdefinition(self.defn_or_decl):
            self.defn_or_decl.wrappedModule = self.coreir_module
            libs = self.backend.included_libs()
            self.defn_or_decl.coreir_wrapped_modules_libs_used = libs


class InstanceTransformer(LeafTransformer):
    def __init__(self, backend, opts, inst, defn):
        super().__init__(backend, opts)
        self.inst = inst
        self.defn = defn
        self.coreir_inst_gen = None

    def run_self(self):
        self.coreir_inst_gen = self.run_self_impl()

    def run_self_impl(self):
        _logger.debug(
            f"Compiling instance {(self.inst.name, type(self.inst).name)}"
        )
        defn = type(self.inst)
        if hasattr(self.inst, "namespace"):
            lib = self.backend.get_lib(self.inst.namespace)
        else:
            lib = self.backend.get_lib(self.inst.coreir_lib)
            if self.inst.coreir_lib == "global":
                lib = self.opts.get("user_namespace", lib)
        if self.inst.coreir_genargs is None:
            module = get_module_of_inst(self.backend.context, self.inst, lib)
            args = get_inst_args(self.inst)
            args = self.backend.context.new_values(args)
            return lambda m: m.add_module_instance(self.inst.name, module, args)
        generator = lib.generators[defn.coreir_name]
        config_args = {k: v for k, v in self.inst.coreir_configargs.items()}
        config_args = self.backend.context.new_values(config_args)
        gen_args = {k: map_genarg(self.backend.context, v)
                    for k, v in defn.coreir_genargs.items()}
        gen_args = self.backend.context.new_values(gen_args)
        return lambda m: m.add_generator_instance(
            self.inst.name, generator, gen_args, config_args)


class WrappedTransformer(LeafTransformer):
    def __init__(self, backend, opts, defn):
        super().__init__(backend, opts)
        self.defn = defn
        self.coreir_module = self.defn.wrappedModule
        self.backend.include_lib_or_libs(
            self.defn.coreir_wrapped_modules_libs_used)


class DefinitionTransformer(TransformerBase):
    def __init__(self, backend, opts, defn):
        super().__init__(backend, opts)
        self.defn = defn
        self.coreir_module = None
        self.decl_tx = DeclarationTransformer(self.backend,
                                              self.opts,
                                              self.defn)
        self.inst_txs = {
            inst: InstanceTransformer(self.backend, self.opts, inst, self.defn)
            for inst in self.defn.instances
        }
        self.clocks = get_default_clocks(defn)
        self._constant_cache = {}

    def children(self):
        pass_ = InstanceGraphPass(self.defn)
        pass_.run()
        deps = [k for k, _ in pass_.tsortedgraph if k is not self.defn]
        children = [DefnOrDeclTransformer(self.backend, self.opts, dep)
                    for dep in deps]
        children += [self.decl_tx]
        children += self.inst_txs.values()
        return children

    def run_self(self):
        _logger.debug(f"Compiling definition {self.defn}")
        self.coreir_module = self.decl_tx.coreir_module
        if self.defn.inline_verilog_strs:
            inline_verilog = "\n\n".join(x[0] for x in
                                         self.defn.inline_verilog_strs)
            connect_references = {}
            for _, inline_value_map in self.defn.inline_verilog_strs:
                for key, value in inline_value_map.items():
                    connect_references[key] = magma_port_to_coreir_port(value)
            self.coreir_module.add_metadata("inline_verilog", json.dumps(
                {"str": inline_verilog,
                 "connect_references": connect_references}
            ))
        for name, module in self.defn.compiled_bind_modules.items():
            self.backend.bind_module(name, module)

        self.coreir_module.definition = self.get_coreir_defn()

    def get_coreir_defn(self):
        coreir_defn = self.coreir_module.new_definition()
        coreir_insts = {inst: self.inst_txs[inst].coreir_inst_gen(coreir_defn)
                        for inst in self.defn.instances}
        # If this module was imported from verilog, do not go through the
        # general module construction flow. Instead just attach the verilog
        # source as metadata and return the module.
        if hasattr(self.defn, "verilogFile") and self.defn.verilogFile:
            metadata = json.dumps({"verilog_string": self.defn.verilogFile})
            self.coreir_module.add_metadata("verilog", metadata)
            return coreir_defn
        if self.defn.coreir_lib is not None:
            self.backend.include_lib_or_libs(self.defn.coreir_lib)
        for name, port in self.defn.interface.ports.items():
            _logger.debug(f"{name}, {port}, {port.is_output()}")
        for inst, coreir_inst in coreir_insts.items():
            if get_codegen_debug_info() and getattr(inst, "debug_info", False):
                attach_debug_info(coreir_inst, inst.debug_info)
            if getattr(inst, "coreir_metadata"):
                for k, v in inst.coreir_metadata.items():
                    coreir_inst.add_metadata(k, json.dumps(v))
        for inst in coreir_insts:
            for name, port in inst.interface.ports.items():
                self.connect_non_outputs(coreir_defn, port)
        for port in self.defn.interface.ports.values():
            self.connect_non_outputs(coreir_defn, port)
        return coreir_defn

    def connect_non_outputs(self, module_defn, port):
        # Recurse into non input types that may contain inout children.
        if isinstance(port, Tuple) and not port.is_input() or \
           isinstance(port, Array) and not port.T.is_input():
            for elem in port:
                self.connect_non_outputs(module_defn, elem)
        elif not port.is_output():
            self.connect(module_defn, port, port.trace())

    def get_source(self, port, value, module_defn):
        port = _unwrap(port)
        value = _unwrap(value)
        if isinstance(value, Wireable):
            return value
        if isinstance(value, Slice):
            return module_defn.select(value.get_coreir_select())
        if isinstance(value, Bits) and value.const():
            return self._const_instance(value, len(value), module_defn)
        if value.anon() and isinstance(value, Array):
            drivers = _collect_drivers(value)
            offset = 0
            for d in drivers:
                d = _unwrap(d)
                if len(d) == 1:
                    # _collect_drivers will introduce a slice of length 1 for
                    # non-slices, so we index them here with 0 to unpack the
                    # extra array dimension
                    self.connect(module_defn, port[offset], d[0])
                else:
                    self.connect(module_defn,
                                 Slice(port, offset, offset + len(d)),
                                 Slice(d[0].name.array, d[0].name.index,
                                       d[-1].name.index + 1))
                offset += len(d)

            return None
        if isinstance(value, Tuple) and value.anon():
            for p, v in zip(port, value):
                self.connect(module_defn, p, v)
            return None
        if value.const():
            return self._const_instance(value, None, module_defn)
        if isinstance(value.name, PortViewRef):
            return module_defn.select(
                magma_name_to_coreir_select(value.name))
        return module_defn.select(magma_port_to_coreir_port(value))

    def connect(self, module_defn, port, value):
        if value is None and is_clock_or_nested_clock(type(port)):
            if not wire_default_clock(port, self.clocks):
                # No default clock
                return
            value = port.trace()
        if value is None:
            if port.is_inout():
                return  # skip inouts because they might be conn. as an input.
            if getattr(self.defn, "_ignore_undriven_", False):
                return
            error_str = f"Found unconnected port: {port.debug_name}\n"
            error_str += _make_unconnected_error_str(port)
            raise Exception(error_str)
        source = self.get_source(port, value, module_defn)
        if not source:
            return
        sink = module_defn.select(magma_port_to_coreir_port(port))
        module_defn.connect(source, sink)
        if get_codegen_debug_info() and getattr(port, "debug_info", False):
            attach_debug_info(module_defn, port.debug_info, source, sink)

    def _const_instance(self, constant, num_bits, module_defn):
        value = constant_to_value(constant)
        key = (value, num_bits)
        try:
            return self._constant_cache[key]
        except KeyError:
            pass
        if num_bits is None:
            config = self.backend.context.new_values({"value": bool(value)})
            name = f"bit_const_{value}_{num_bits}"
            mod = self.backend.get_lib("corebit").modules["const"]
            module_defn.add_module_instance(name, mod, config)
        else:
            config = self.backend.context.new_values({"value": value})
            name = f"const_{value}_{num_bits}"
            gen = self.backend.get_lib("coreir").generators["const"]
            gen_args = self.backend.context.new_values({"width": num_bits})
            module_defn.add_generator_instance(name, gen, gen_args, config)
        out = module_defn.select(f"{name}.out")
        return self._constant_cache.setdefault(key, out)


class DeclarationTransformer(LeafTransformer):
    def __init__(self, backend, opts, decl):
        super().__init__(backend, opts)
        self.decl = decl
        self.coreir_module = None

    def run_self(self):
        self.coreir_module = self.run_self_impl()

    def run_self_impl(self):
        self.decl = self.decl
        _logger.debug(f"Compiling declaration {self.decl}")
        if self.decl.coreir_lib is not None:
            self.backend.include_lib_or_libs(self.decl.coreir_lib)
        # These libraries are already available by default in coreir, so we
        # don't need declarations.
        if self.decl.coreir_lib in ["coreir", "corebit", "commonlib",
                                    "memory"]:
            lib = self.backend.get_lib(self.decl.coreir_lib)
            if self.decl.coreir_genargs is None:
                return lib.modules[self.decl.coreir_name]
            return lib.generators[self.decl.coreir_name]
        try:
            coreir_module = self.backend.get_module(self.decl)
            _logger.debug(f"{self.decl} already compiled, skipping")
            return coreir_module
        except KeyError:
            pass
        if get_debug_mode():
            check_magma_interface(self.decl.interface)
        module_type = magma_interface_to_coreir_module_type(
            self.backend.context, self.decl.interface)
        if isinstance(self.decl.interface, InterfaceKind):
            module_type = self.backend.context.Flip(module_type)
        kwargs = {}
        if hasattr(self.decl, "coreir_config_param_types"):
            param_types = self.decl.coreir_config_param_types
            kwargs["cparams"] = make_cparams(self.backend.context, param_types)
        if hasattr(self.decl, "namespace"):
            # Allows users to choose namespace explicitly with
            # class MyCircuit(m.Circuit):
            #     namespace = "foo"
            # overrides user_namespace setting
            namespace = self.backend.get_lib(self.decl.namespace)
        else:
            namespace = self.opts.get("user_namespace",
                                      self.backend.context.global_namespace)
        coreir_module = namespace.new_module(
            self.decl.coreir_name, module_type, **kwargs)
        if get_codegen_debug_info() and self.decl.debug_info:
            attach_debug_info(coreir_module, self.decl.debug_info)
        for key, value in self.decl.coreir_metadata.items():
            coreir_module.add_metadata(key, json.dumps(value))
        return coreir_module



from magma.array import Array
from magma.circuit import Circuit
from magma.digital import Digital
from magma.ref import (DefnRef, InstRef, NamedRef, ArrayRef, TupleRef,
                       PortViewRef)
from magma.t import Type, Direction
from magma.tuple import Tuple
from magma.view import InstView, PortView


def is_nd_array(T, skip=True):
    if issubclass(T, Digital) and not skip:
        return True
    if issubclass(T, Array):
        return is_nd_array(T.T, False)
    return False


def _should_recurse(value, disable_ndarray):
    if isinstance(value, PortView):
        value = value.port
    return (isinstance(value, Array) and not issubclass(value.T, Digital) and
            is_nd_array(type(value)) and disable_ndarray)


def value_to_verilog_name(value, disable_ndarray=False):
    if _should_recurse(value, disable_ndarray):
        elems = ", ".join(value_to_verilog_name(t, disable_ndarray)
                          for t in reversed(value))
        return f"'{{{elems}}}"
    elif isinstance(value, Tuple):
        raise NotImplementedError("Inlining unflattened tuple")
    elif isinstance(value, InstView):
        prefix = ""
        if value.parent is not None:
            prefix = value_to_verilog_name(value.parent, disable_ndarray) + "."
        return prefix + value.inst.name
    return verilog_name(value.name, disable_ndarray=disable_ndarray)


def verilog_name(name, inst_sep="_", disable_ndarray=False):
    if isinstance(name, PortViewRef):
        curr = name.view.parent
        hierarchical_path = curr.inst.name + "."
        while isinstance(curr.parent, InstView):
            hierarchical_path = curr.parent.inst.name + "." + hierarchical_path
            curr = curr.parent
        return hierarchical_path + verilog_name(
            name.view.port.name, disable_ndarray=disable_ndarray)
    if isinstance(name, DefnRef):
        return str(name)
    if isinstance(name, InstRef):
        return f"{name.inst.name}{inst_sep}{str(name)}"
    if isinstance(name, NamedRef):
        return str(name)
    if isinstance(name, ArrayRef):
        array_name = verilog_name(name.array.name,
                                  disable_ndarray=disable_ndarray)
        if (issubclass(name.array.T, Digital) or
                is_nd_array(type(name.array)) and not disable_ndarray):
            index = f"[{name.index}]"
        else:
            index = f"_{name.index}"
        return f"{array_name}{index}"
    if isinstance(name, TupleRef):
        tuple_name = verilog_name(name.tuple.name,
                                  disable_ndarray=disable_ndarray)
        index = name.index
        if name.root() is None:
            # Not a named temporary that will be flattened to a wire
            try:
                int(index)
                # python/coreir don't allow pure integer names
                index = f"_{index}"
            except ValueError:
                pass
        return f"{tuple_name}_{index}"
    raise NotImplementedError(name, type(name))



import inspect
import sys
import re
import ast


# The code in this file is adapted from the Silica compiler, shout out to Teguh
# for the help on the original implementation!
# Automatically generate is_{} functions for classes in the ast module
module_obj = sys.modules[__name__]

_first_cap_re = re.compile(r'(.)([A-Z][a-z]+)')
_all_cap_re = re.compile(r'([a-z0-9])([A-Z])')


def to_snake_case(name):
    """ Converts name from CamelCase to snake_case

    https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    """
    s1 = re.sub(_first_cap_re, r'\1_\2', name)
    return re.sub(_all_cap_re, r'\1_\2', s1).lower()


def is_generator(name):
    def f(node):
        return isinstance(node, getattr(ast, name))

    setattr(module_obj, "is_" + to_snake_case(name), f)


if sys.version_info < (3, 9):
    for x in (m[0] for m in inspect.getmembers(ast, inspect.isclass) if
              m[1].__module__ == '_ast'):
        is_generator(x)
else:
    for x in (m[0] for m in inspect.getmembers(ast, inspect.isclass) if
              m[1].__module__ == 'ast'):
        is_generator(x)


def get_width(node, width_table, defn_env={}):
    if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and \
       node.func.id in {"bits", "uint", "BitVector"}:
        if isinstance(node.args[1], ast.Call) and \
                node.args[1].func.id == "eval":
            return eval(astor.to_source(node.args[1]).rstrip())
        if not isinstance(node.args[1], ast.Num):
            raise TypeError(f"Cannot get width of "
                            f"{astor.to_source(node.args[1]).rstrip()}, we "
                            f"should know all widths at compile time.")
        return node.args[1].n
    if isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and \
       node.func.id in {"bit"}:
        return None
    elif isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and \
            node.func.id in {"memory"}:
        return MemoryType(node.args[0].n, node.args[1].n)
    elif isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and \
            node.func.id in {"zext"}:
        assert isinstance(node.args[1], ast.Num), \
            "We should know all widths at compile time"
        return get_width(node.args[0], width_table) + node.args[1].n
    elif isinstance(node, ast.Name):
        if node.id not in width_table:
            raise Exception(f"Trying to get width of variable that hasn't "
                            f"been previously added to the width_table: "
                            f"{node.id} (width_table={width_table})")
        return width_table[node.id]
    elif isinstance(node, ast.Call):
        if isinstance(node.func, ast.Name):
            widths = [get_width(arg, width_table) for arg in node.args]
            if node.func.id in {"add", "xor", "sub", "and_", "not_"}:
                if not all(widths[0] == x for x in widths):
                    raise TypeError(f"Calling {node.func.id} with different "
                                    "length types")
                width = widths[0]
                if node.func.id == "add":
                    for keyword in node.keywords:
                        if keyword.arg == "cout" and \
                                isinstance(keyword.value, ast.NameConstant) \
                                and keyword.value.value is True:
                            width = (width, None)
                return width
            elif node.func.id == "eq":
                if not all(widths[0] == x for x in widths):
                    raise TypeError("Calling eq with different length types")
                return None
            elif node.func.id == "decoder":
                return 2 ** get_width(node.args[0], width_table)
            else:
                raise NotImplementedError(ast.dump(node))
        else:
            raise NotImplementedError(ast.dump(node))
    elif isinstance(node, ast.UnaryOp):
        return get_width(node.operand, width_table)
    elif isinstance(node, ast.BinOp):
        left_width = get_width(node.left, width_table)
        right_width = get_width(node.right, width_table)
        if left_width != right_width:
            raise TypeError(f"Binary operation with mismatched widths "
                            f"({left_width}, {right_width}) {ast.dump(node)}")
        return left_width
    elif isinstance(node, ast.IfExp):
        left_width = get_width(node.body, width_table)
        if node.orelse:
            right_width = get_width(node.orelse, width_table)
            if left_width != right_width:
                raise TypeError(f"Binary operation with mismatched widths "
                                f"{ast.dump(node)}")
        return left_width
    elif isinstance(node, ast.Compare):
        # TODO: Check widths of operands
        return None
    elif isinstance(node, ast.NameConstant) and node.value in [True, False]:
        return None
    elif isinstance(node, ast.List):
        widths = [get_width(arg, width_table) for arg in node.elts]
        assert all(widths[0] == width for width in widths)
        return MemoryType(len(node.elts), widths[0])
    elif isinstance(node, ast.Subscript):
        if isinstance(node.slice, ast.Index):
            width = get_width(node.value, width_table)
            if isinstance(width, tuple):
                return width[1]
            elif isinstance(width, MemoryType):
                return width.width
            return None
        elif isinstance(node.slice, ast.Slice):
            width = get_width(node.value, width_table)
            if node.slice.lower is None and isinstance(node.slice.upper,
                                                       ast.Num):
                if node.slice.step is not None:
                    raise NotImplementedError()
                return node.slice.upper.n
            elif node.slice.upper is None:
                lower = 0
                if isinstance(node.slice.lower, ast.Num):
                    lower = node.slice.lower.n
                width = width_table[node.value.id]
                if isinstance(width, MemoryType):
                    return MemoryType(width.height - lower, width.width)
                else:
                    return width - lower
            elif isinstance(node.slice.upper, ast.Num):
                lower = node.slice.lower.n
                upper = node.slice.upper.n
                width = width_table[node.value.id]
                if isinstance(width, MemoryType):
                    return MemoryType(upper - lower - 1, width.width)
                else:
                    return upper - lower - 1
    elif isinstance(node, ast.Num):
        return max(node.n.bit_length(), 1)
    elif isinstance(node, ast.Attribute):
        type_ = width_table[node.value.id]._outputs[node.attr]
        if isinstance(type_, m.BitKind):
            return None
        raise NotImplementedError(type_)

    raise NotImplementedError(ast.dump(node))


class CollectInitialWidthsAndTypes(ast.NodeVisitor):
    def __init__(self, width_table, type_table, defn_env):
        self.width_table = width_table
        self.type_table = type_table
        self.defn_env = defn_env

    def visit_Assign(self, node):
        if len(node.targets) == 1:
            if isinstance(node.targets[0], ast.Name):
                if isinstance(node.value, ast.Yield):
                    pass  # width specified at compile time
                elif isinstance(node.value, ast.Call) and \
                        isinstance(node.value.func, ast.Attribute) and \
                        node.value.func.attr == "send":
                    pass
                elif node.targets[0].id not in self.width_table:
                    self.width_table[node.targets[0].id] = \
                        get_width(node.value, self.width_table, self.defn_env)
                    if isinstance(node.value, ast.Call) and \
                            isinstance(node.value.func, ast.Name) and \
                            node.value.func.id in {"bits", "uint", "bit"}:
                        self.type_table[node.targets[0].id] = node.value.func.id
                    elif isinstance(node.value, ast.NameConstant) and \
                            node.value.value in [True, False]:
                        self.type_table[node.targets[0].id] = "bit"
                    elif isinstance(node.value, ast.Name) and \
                            node.value.id in self.type_table:
                        self.type_table[node.targets[0].id] = \
                            self.type_table[node.value.id]


class PromoteWidths(ast.NodeTransformer):
    def __init__(self, width_table, type_table):
        self.width_table = width_table
        self.type_table = type_table

    def check_valid(self, int_length, expected_length):
        if expected_length is None and int_length <= 1:
            return
        if int_length <= expected_length:
            return
        raise TypeError("Cannot promote integer with greater width than other "
                        "operand")

    def make(self, value, width, type_):
        if type_ == "bit":
            return ast.parse(f"{type_}({value})").body[0].value
        else:
            return ast.parse(f"{type_}({value}, {width})").body[0].value

    def get_type(self, node):
        if isinstance(node, ast.Name):
            return self.type_table[node.id]
        elif isinstance(node, ast.Subscript):
            if isinstance(node.value, ast.Name):
                type_ = self.type_table[node.value.id]
                if type_ == "uint" and isinstance(node.slice, ast.Index):
                    return "bit"
                elif type_ == "uint" and isinstance(node.slice, ast.Slice):
                    if node.slice.lower is None and \
                            isinstance(node.slice.upper, ast.Num):
                        if node.slice.step is not None:
                            raise NotImplementedError()
                        # return node.slice.upper.n
                        return "bits"
                else:
                    raise NotImplementedError(ast.dump(node))
        elif isinstance(node, ast.BinOp):
            left_type = self.get_type(node.left)
            right_type = self.get_type(node.right)
            assert left_type == right_type, (left_type, right_type)
            return left_type
        elif isinstance(node, ast.Call) and isinstance(node.func, ast.Name) \
                and node.func.id in ["bits"]:
            return node.func.id
        raise NotImplementedError(ast.dump(node))

    def visit_Assign(self, node):
        node.value = self.visit(node.value)
        if isinstance(node.value, ast.Num):
            width = get_width(node.targets[0], self.width_table)
            self.check_valid(node.value.n.bit_length(), width)
            node.value = self.make(node.value.n, width,
                                   self.get_type(node.targets[0]))
        elif isinstance(node.value, ast.NameConstant):
            width = get_width(node.targets[0], self.width_table)
            node.value = ast.parse(f"bit({node.value.value})").body[0].value
        return node

    def visit_AugAssign(self, node):
        node.value = self.visit(node.value)
        if isinstance(node.value, ast.Num):
            width = get_width(node.target, self.width_table)
            self.check_valid(node.value.n.bit_length(), width)
            node.value = self.make(node.value.n, width,
                                   self.get_type(node.target))
        return node

    def visit_BinOp(self, node):
        node.left = self.visit(node.left)
        node.right = self.visit(node.right)
        if isinstance(node.left, ast.Num):
            right_width = get_width(node.right, self.width_table)
            try:
                self.check_valid(node.left.n.bit_length(), right_width)
            except TypeError as e:
                print(f"Error type checking node {ast.dump(node)}")
                raise e
            node.left = self.make(node.left.n, right_width,
                                  self.get_type(node.right))
        elif isinstance(node.right, ast.Num):
            left_width = get_width(node.left, self.width_table)
            try:
                self.check_valid(node.right.n.bit_length(), left_width)
            except TypeError as e:
                print(f"Error type checking node {ast.dump(node)}")
                raise e
            node.right = self.make(node.right.n, left_width,
                                   self.get_type(node.left))
        return node

    def visit_Compare(self, node):
        node.left = self.visit(node.left)
        node.comparators = [self.visit(x) for x in node.comparators]
        if not isinstance(node.left, ast.Num):
            left_width = get_width(node.left, self.width_table)
            type_ = self.get_type(node.left)
            for i in range(len(node.comparators)):
                if isinstance(node.comparators[i], ast.Num):
                    self.check_valid(node.comparators[i].n.bit_length(),
                                     left_width)
                    node.comparators[i] = self.make(node.comparators[i].n,
                                                    left_width, type_)
        else:
            for comparator in node.comparators:
                if not isinstance(comparator, ast.Num):
                    width = get_width(comparator, self.width_table)
                    type_ = self.get_type(comparator)
                    break
            else:
                assert False, \
                    "Constant fold should have folded this expression " \
                    f"{ast.dump(node)}"
            self.check_valid(node.left.n.bit_length(), width)
            node.left = self.make(node.left.n, width, type_)
            for i in range(len(node.comparators)):
                if isinstance(node.comparators[i], ast.Num):
                    self.check_valid(node.comparators[i].n.bit_length(), width)
                    node.comparators[i] = self.make(node.comparators[i].n,
                                                    width, type_)
        return node


class RemoveBits(ast.NodeTransformer):
    def visit_Call(self, node):
        if is_attribute(node.func) and \
                is_name(node.func.value) and node.func.value.id == "m" and \
                node.func.attr == "bits":
            return node.args[0]
        return node
